package de.fitko.fitconnect.client.di;

import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.data.Data;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionRequest;
import de.fitko.fitconnect.api.domain.model.submission.CreatedSubmission;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.client.ClientNotAuthenticated;
import de.fitko.fitconnect.api.services.Sender;

import java.io.File;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Collectors;

/**
 * A fluent client for announcing and handing in a {@link SubmissionRequest}
 */
public class FluentSenderClient extends Client {

    private FluentSenderClient() {}

    public static WithDestination build() {
        final Sender sender = getService(Sender.class);
         final String clientId = getProperty("clientId");
         final String secret = getProperty("clientSecret");
        return new ClientBuilder(sender, clientId, secret);
    }

    public interface WithDestination {

        /**
         * Configures the client for the given destination and load the public key
         *
         * @param destinationId the clients destination
         * @return the upload step for attachments
         */
        WithAttachments withDestination(UUID destinationId);

        CreatedSubmission send(UUID destinationId, List<File> attachments, String data);

        CreatedSubmission send(UUID destinationId, List<File> attachments, byte[] data);

        CreatedSubmission send(UUID destinationId, List<File> attachments);
    }

    public interface WithAttachments {

        /**
         * Sends the submission with a list of attachments
         *
         * @param attachments that are sent with the submission
         * @return the step where additional data can be added to the submission
         */
        WithData withAttachments(List<File> attachments);
    }

    public interface WithData {

        Submit withData(String data);

        Submit withData(byte[] data);

        CreatedSubmission submit();
    }

    public interface Submit {
        CreatedSubmission submit();
    }

    public static class ClientBuilder implements WithDestination, WithAttachments, WithData, Submit {

        private Sender sender;
        private final String clientId;
        private final String secret;

        private UUID destinationId;
        private Data data;
        private List<Attachment> attachments;
        private RSAKey encryptionKey;
        private UUID submissionId;

        public ClientBuilder(Sender sender, String clientId, String secret) {
            this.sender = sender;
            this.clientId = clientId;
            this.secret = secret;
            authenticate();
        }

        @Override
        public WithAttachments withDestination(UUID destinationId) {
            final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId);
            final ValidationResult validationResult = sender.validatePublicKey(encryptionKey);
            if(validationResult.hasErrors()){
                throw new RuntimeException("Public encryption key is not valid");
            }
            this.encryptionKey = encryptionKey;
            return this;
        }

        @Override
        public CreatedSubmission send(UUID destinationId, List<File> attachments, String data) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .withData(data)
                    .submit();
        }

        @Override
        public CreatedSubmission send(UUID destinationId, List<File> attachments, byte[] data) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .withData(data)
                    .submit();
        }

        @Override
        public CreatedSubmission send(UUID destinationId, List<File> attachments) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .submit();
        }

        @Override
        public Submit withData(String data) {
            return withData(data.getBytes(StandardCharsets.UTF_8));
        }

        @Override
        public Submit withData(byte[] data) {
            var unencryptedData = Data.builder().build();
            final Optional<Data> encryptedData = sender.encryptSubmissionData(encryptionKey, unencryptedData);
            if(encryptedData.isEmpty()){
                throw new RuntimeException("Data could not be encrypted");
            }
            this.data = encryptedData.get();
            return this;
        }

        @Override
        public WithData withAttachments(List<File> attachmentFiles) {
            final List<Attachment> attachments = readFilesToAttachments(attachmentFiles);
            final CreatedSubmission submission = createSubmission(destinationId, attachments);
            this.submissionId = submission.getSubmissionId();
            final List<Attachment> encryptedAttachments = encryptAttachments(encryptionKey, attachments);
            sender.uploadAttachments(submissionId, encryptedAttachments);
            return this;
        }

        @Override
        public CreatedSubmission submit() {
            final Optional<Metadata> metadata = getMetadata();
            if(metadata.isEmpty()){
                throw new RuntimeException("Metadata could not be created");
            }
            return sender.sendSubmission(submissionId, metadata.get()).orElseThrow();
        }

        private Optional<Metadata> getMetadata() {
            return this.data == null ? sender.createMetadata(attachments) : sender.createMetadata(data, attachments);
        }

        private List<Attachment> encryptAttachments(RSAKey encryptionKey, List<Attachment> attachments) {
            return attachments.stream().map(attachment -> encryptAttachment(encryptionKey, attachment)).collect(Collectors.toList());
        }

        private Attachment encryptAttachment(RSAKey encryptionKey, Attachment attachment) {
            final Optional<Attachment> encryptedAttachment = sender.encryptAttachment(encryptionKey, attachment);
            if (encryptedAttachment.isEmpty()) {
                throw new RuntimeException("Attachment could not be encrypted");
            }
            return encryptedAttachment.get();
        }

        private CreatedSubmission createSubmission(UUID destinationId, List<Attachment> attachments) {
            final SubmissionRequest request = SubmissionRequest.builder()
                    .destinationId(destinationId)
                    .announcedAttachments(asListOfAttachmentsIds(attachments))
                    .build();

            final Optional<CreatedSubmission> response = sender.createSubmission(request);
            if (response.isEmpty()) {
                throw new RuntimeException("Could not create submission");
            }
            return response.get();
        }

        private List<UUID> asListOfAttachmentsIds(List<Attachment> attachments) {
            return attachments.stream().map(attachment -> attachment.getAttachmentId()).collect(Collectors.toList());
        }

        private void authenticate() {
            final Optional<OAuthToken> oAuthToken = sender.retrieveOAuthToken(clientId, secret);
            if (oAuthToken.isEmpty()) {
                throw new ClientNotAuthenticated("Client is not authenticated, please authenticate first");
            }
        }

        private List<Attachment> readFilesToAttachments(List<File> attachmentFiles) {
            return attachmentFiles.stream()
                    .map(this::readFileToAttachment)
                    .collect(Collectors.toList());
        }

        private Attachment readFileToAttachment(File file) {
            final Optional<Attachment> attachment = sender.createAttachment(file);
            if(attachment.isEmpty()){
                throw new RuntimeException("Attachment could not be created");
            }
            return attachment.get();
        }
    }
}